#!/usr/bin/perl
# 2013-11-18 / dosman / Home Heartbeat script - It digests the state table of the Home Heartbeat base station
# v1.0                  and outputs human readable data. I have not tested the email functionality yet, but the 
#                       other functions work fine. For updates and other info about this script check out this page:
#                       http://bloominglabs.org/index.php/Eaton_Home_Heartbeat
#
#####################################################################################################################
# The Vendor ID for the FTDI chip in the base station is unique to Eaton, but otherwise this is a genric FTDI USB->Serial interface.
# Add a modprobe line for ftdi_sio driver, get vendor and product lines from lsusb:
# echo "options ftdi_sio vendor=0x403 product=0xde29" > /etc/modprobe.d/ftdi_sio.conf
#
# Reload ftdi_sio (or reboot) and it should be picked up correctly now:
# modprobe -r ftdi_sio
# modprobe ftdi_sio
#
#####################################################################################################################
# All field information obtained from this site and my own hardware:
# http://www.kolinahr.com/documentation/home-heartbeat/state-table-of-the-eaton-home-heartbeat/
#
# Output field format:
# STATE="F1,F2,F3,F4,F5,F6,F7,F8,F9,F10,F11,F12,F13,F14,F15,F16,F17"
# STATE="02,00,0040,0002,00,12,00,00,0301,12,0000,00,FF,00000000,00,000D6F0000011367,Home Key"


# Field 1 State Record ID 0 This is a number associated with each device as it's associated with the base station, however it can be changed so do not rely on 
# this. Using "Forget Device" on the home key will cause the devices to renumber.

# Field 2 ZigBee Binding ID 0 This value is an index into the ZigBee Binding Table.

# Field 3 Device Capabilities - field is constant for each device, this field is still not fully understood
$CAPABILITY_UNKNOWN1="01";
$CAPABILITY_UNKNOWN2="02";
$CAPABILITY_SENSOR1="04";
$CAPABILITY_BASESTATION="08";
$CAPABILITY_RECEIVESPARAMS="10"; # Motion, Power, and Reminder sensors
$CAPABILITY_SENSOR2="20";
$CAPABILITY_HOMEKEY="40";
$CAPABILITY_AC="80"; # Base station, modem, and power sensors

# Field 4 Device Type
$TYPE_BASE="0001";
$TYPE_HOMEKEY="0002";
$TYPE_OPENCLOSE="0003";
$TYPE_POWER="0004";
$TYPE_WATER="0005";
$TYPE_REMINDER="0006";
$TYPE_ATTENTION="0007";
$TYPE_MODEM="0010";
$TYPE_MOTION="0017";
$TYPE_TILT="0018";

# Field 5 Device State - This field contains the last reported state of the device.
# Field 6 Device State Timer - This is the elapsed time since the device state has changed.

# Field 7 Device Alerts
$ALARM_TRIGGERED="01";
$DEV_OFFLINE="02"; # device offline, out of range, or battery dead
$BATTERY_LOW="04";
$BATTERY_CHARGING="08"; # only used by home key
$RUNNING_ON_BATTERY="20"; # only used by base station

# Field 8 Device Name Index - This is an index into a predefined list of available names specific to the device type, also Field 17 contains the complete name retrieved out of the index.

# Field 9 Device Configuration
$ALARM1="0001";
$ALARM2="0002";
$CALLME1="0100";
$CALLME2="0200";

# Field 10 Alive Update Timer - This is the elapsed time since that last “I’m Alive” message received from the device.
# Field 11 Update Flags - This field is normally zero (0x00) for all devices. It has been observed to contain the value 0x80 when there is pending data needing to be sent to the device. Please read the description under Field 13: Device Parameter section below.
# Field 12 Undefined
# Field 13 Device Parameter - This field may contain a user selectable configuration value. Only a few devices have the ability to accept a configuration parameter. Power, Reminder, and Motion.
# Field 14 Undefined
# Field 15 Pending Update Timer - This is the elapsed time since the user changed the device configuration parameter defined in Field 13
# Field 16 MAC Address
# Field 17 Device Name

#####################################################################################################################

use Text::Table;
use Getopt::Std;

$errorLog = "/var/log/hh.log";
$alertLog = "/var/log/hh.log";
$htmlOut="/var/www/html/sensors.html";
$fontSize = "6"; # For HTML output on small screens, set font size big

# You can choose what alarm types to be emailed about
# values are or'ed  together
#$alert = "00001"; # Notify on alarm triggered
#$alert = "00010"; # Notify on low battery
#$alert = "00100"; # Notify on battery charging / only homekey does this / why would you care?
#$alert = "01000"; # Notify on base station running on backup battery
#$alert = "10000"; # Notify on any unknown alert types / this would be caused by an alert unknown to this script or a script bug
$alert = "01001"; # Notify on alartm triggered and base station running on backup battery
$emailFrom = 'homeheartbeat@localhost';
$emailTo = 'user@localhost';


# Only 1 of theese 2 methods should be uncommented, and the chosen method needs to be customized with appropriate values
&useSerial();
#&useTelnet();

###################################################################################################################
open ERRORLOG, ">>$errorLog" or die $!;
open ALERTLOG, ">>$alertLog" or die $!;

use POSIX qw(tzset);
$ENV{TZ} = 'America/New_York'; # Set your timezone here...
tzset;
$date = localtime;

# This is handy to use with a Twistd deamon when connecting our base station to a TCP port...
sub useTelnet
{
my $host = "localhost";
my $port = "4001";

use Net::Telnet;
  $telnet = new Net::Telnet (Timeout=>10, Port=>$port,
  Errmode=>'die');

  $telnet->open("$host");

  # Request the state table
  $telnet->print('S');

  while (1)
  {
    my $input = $telnet->getline();
    $input =~ s/\r//g;
    if ($input ne ""){
        push(@DEVICES, "$input");
    }
    if ($input =~ /STATE=DONE/){
      pop(@DEVICES);
      last;
    }
    $input = "";
  }
  
  #$output4 = $telnet->getline();
}
#telnetData;

# Most people will probably want to use this instead for direct USB-serial connectivity
sub useSerial
{
  $DEV="/dev/ttyUSB0";

  @DEVICES=();
  use Device::SerialPort;
  $port = Device::SerialPort->new("$DEV");
  $port->databits(8);
  $port->baudrate(38400); # All Home-Heartbeat base stations should use this baud rate
  $port->parity("none");
  $port->stopbits(1);
  
  # Request the state table
  $port->write("S\n");

  while (1)
  {
    my $input = $port->lookfor();
    $input =~ s/\r//g;
    if ($input ne ""){
        push(@DEVICES, "$input");
    }
    if ($input =~ /STATE=DONE/){
      # Dump the STATE=DONE line
      pop(@DEVICES);
      last;
    }
    $input = "";
  }
}

#####################################################################################################################

my %args;
getopts('rewhd', \%args);
if($args{r}){
  $flag = "r";
  while(scalar(@DEVICES) > 0) {
   my $i = shift(@DEVICES);
   print $i, "\n"; 
  }
  exit;
} elsif ($args{e}){ # Extended output format
    $flag = "e";
    $tb = Text::Table->new(
        "Device Type", "Custom Name", "Last Event", "Last Seen", "Alert Status", "Configuration", "Record ID", "Capability ID", "Name Index", "Update Flag", "Device Parameter", "Pending Update Timer", "ZigBee Binding ID", "ZigBee Mac"
    );
} elsif ($args{w}){ # Web/HTML output
    $flag = "w";
} elsif ($args{d}){ # debug output / extra information about the system
    print "This doesn't do anything yet\n";
    $flag = "d";
    exit;
 
#    $port->write("S\n");
#    while (1)
#    {
#      my $input = $port->lookfor();
#      $input =~ s/\r//g;
#      if ($input ne ""){
#          push(@DEVICES, "$input");
#      }
#      if ($input =~ /STATE=DONE/){
#        # Dump the STATE=DONE line
#        pop(@DEVICES);
#        last;
#      }
#      $input = "";
#    }

} elsif ($args{h}){
    print "- Home Heartbeat Interface -\n";
    print "This reads the state table from the base station and converts it to human readable format.\n";
    print "This makes it easy to see what the sensors are saying or to more easily feed the data into other scripts.\n";
    print "\nFlags:\n";
    print "        - no flag just gives the most useful output of the state table\n";
    print "-w      - outputs web/html to a file\n";
    print "-r      - just shows the raw state table\n";
    print "-e      - gives an extended output of the state table with more data\n";
#    print "-d      - gives a dump of a number of extra commands from the system\n";
    print "-h      - this help screen\n";
    exit;
} else { # default output format
    $tb = Text::Table->new(
        "Device Type", "Custom Name", "Last Event", "Last Seen", "Alert Status", "Configuration"
    );
}

#####################################################################################################################

while(scalar(@DEVICES) > 0)
{
  my $outp = shift(@DEVICES);
  my ($recordID, $zigBeeID, $devCap, $devType, $devState, $devStateTimer, $devAlert, $devNameIndex, $devConfig, $aliveUpdateTimer, $updateFlags, $undef1, $devParam, $undef2, $pendUpdateTimer, $macAddr, $devName) = split(',', $outp);
  chomp($devName); # fix dev name
  $devName =~ s/"//g;
  $devName =~ s/\r//g;
  my ($dontCare, $recordID) = split('"', $recordID); # fix recordID, this value doesn't have much use though

 # Skip the base station and modem for default and html output. Modem is worthless now, but base station 
 # can be used to detect power outages if host computer is on UPS and batteries are installed in base station.
 if($flag ne "e"){
   if($devType eq $TYPE_MODEM){ 
     next;
   }
   if($devType eq $TYPE_BASE){
     next;
   }
}

# Fill the hopper with data formatted for the appropriate output type
 if ($flag eq "e"){ # extended output
     $tb->load(
       [ &devTypeLookup($devType), "$devName", &decodeTimer($devStateTimer), &decodeTimer($aliveUpdateTimer), &checkAlert(), &devConfigLookup($devType, $devConfig), $recordID, $devCap, $devNameIndex, $updateFlags, $devParam, $pendUpdateTimer, $zigBeeID, $macAddr ]
     );

 } elsif ($flag eq "w"){
    if(! defined $htmlYes){ # Only do this on the first iteration
      &printHtmlHeader();
      $htmlYes = 1;
    }
    print HTMLOUT "<tr><td><font size=$fontSize>$devName</td> <td><font size=$fontSize>";
    print HTMLOUT &decodeTimer($aliveUpdateTimer);
    print HTMLOUT "</td>";

    print HTMLOUT "<td><font size=$fontSize>";
    print HTMLOUT &decodeTimer($devStateTimer);
    print HTMLOUT "</td>";

    print HTMLOUT "<td><font size=$fontSize>";
    print HTMLOUT &checkAlert();
    print HTMLOUT "</td>";

    print HTMLOUT "</tr>\n";

 } else { # default output
   $tb->load(
     [ &devTypeLookup($devType), "$devName", &decodeTimer($devStateTimer), &decodeTimer($aliveUpdateTimer), &checkAlert(), &devConfigLookup($devType, $devConfig ) ]
   );
 }
} # end foreach

if($flag eq "e"){
     print $tb;
} elsif ($flag eq "w"){
     close HTMLOUT;
} else {
     print $tb;
}

close ERRORLOG;
close ALERTLOG;


#############################################################################################
# Begin subroutines

sub printHtmlHeader
{
  #  open(my $output, ">", "/opt/www/house/automation/sensors.html") 
  #       or die "cannot open > sensors.html: $!";
  open HTMLOUT, ">$htmlOut" or die $!;
  print HTMLOUT "<html><title>DOSNET Sensor Network</title><body><meta http-equiv=\"refresh\" content=\"60\"><h1>DOSNET Sensor Network</h1><font size=$fontSize>";
  
  use POSIX qw(tzset);
  $ENV{TZ} = 'America/New_York';tzset;
  my $date = localtime;
  print HTMLOUT "$date";
  print HTMLOUT "<table border=2><tr> <th><font size=$fontSize>Sensor</th> <th><font size=$fontSize>Last seen</th> <th><font size=$fontSize>Last event</th> <th><font size=$fontSize>Alert</th> </tr>\n";
}

sub decodeTimer
{
    my $devStateTimer = hex("@_");
    if($devStateTimer < 64 ){
       return " $devStateTimer secs";
    } elsif ($devStateTimer < 128){
       my $var = $devStateTimer - 64;
       return " $var mins";
    } elsif ($devStateTimer < 160){
       my $var = $devStateTimer - 128;
       return " $var hours";
    } else {
       my $var = $devStateTimer - 160;
       return " $var days";
    }
    # There could be an additional date specification for periods beyond days, we just don't know...
}

sub devTypeLookup
{
  my $devType = $_[0];
  if($devType == $TYPE_BASE){
      return "Base Station";
  } elsif($devType == $TYPE_HOMEKEY){
      return "Home Key";
  } elsif($devType == $TYPE_OPENCLOSE){
      return "Open/Closed";
  } elsif($devType == $TYPE_POWER){
      return "Power Sensor";
  } elsif($devType == $TYPE_WATER){
      return "Water Sensor";
  } elsif ($devType == $TYPE_REMINDER){
      return "Reminder Sensor";
  } elsif ($devType == $TYPE_ATTENTION){
      return "Attention Sensor";
  } elsif ($devType == $TYPE_MODEM){
      return "Modem";
  } elsif ($devType == $TYPE_MOTION){
      return "Motion Sensor";
  } elsif ($devType == $TYPE_TILT){
      return "Tilt Sensor";
  } else {
      print ERRORLOG "$date - devTypeLookup() had unknown devType \"$devType\"\n";
      return "Unknown $devType";
  }
}

sub devConfigLookup
{
  my $devType = $_[0];
  my $devConfig = $_[1];

  if($devType == $TYPE_MOTION){
      if($devConfig == "0000"){
          return "No action set";
      } elsif($devConfig == "0001"){
          return "Alarm on no motion";
      } elsif($devConfig == "0002"){
          return "Alarm on motion";
      } elsif($devConfig == "0100"){
          return "Call on no motion";
      } elsif($devConfig == "0200"){
          return "Call on motion";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_OPENCLOSE){
      if($devConfig == "0001"){
          return "Alarm on closed";
      } elsif($devConfig == "0002"){
          return "Alarm on opened";
      } elsif($devConfig == "0100"){
          return "Call on closed";
      } elsif($devConfig == "0200"){
          return "Call on opened";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_WATER){
      if($devConfig == "0001"){
          return "Alarm on wet";
      } elsif($devConfig == "0002"){
          return "Alarm on dry";
      } elsif($devConfig == "0100"){
          return "Call on wet";
      } elsif($devConfig == "0200"){
          return "Call on dry";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_ATTENION){
      if($devConfig == "0001"){
          return "Alarm on requested";
      } elsif($devConfig == "0002"){
          return "Alarm on standby";
      } elsif($devConfig == "0100"){
          return "Call on requested";
      } elsif($devConfig == "0200"){
          return "Call on standby";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_REMINDER){
      if($devConfig == "0001"){
          return "Alarm on started";
      } elsif($devConfig == "0002"){
          return "Alarm on time's up";
      } elsif($devConfig == "0100"){
          return "Call on started";
      } elsif($devConfig == "0200"){
          return "Call on time's up";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_POWER){
      if($devConfig == "0000"){
          return "No action set";
      } elsif($devConfig == "0001"){
          return "Alarm on Off";
      } elsif($devConfig == "0002"){
          return "Alarm on On";
      } elsif($devConfig == "0100"){
          return "Call on Off";
      } elsif($devConfig == "0200"){
          return "Call on On";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_TILT){
      if($devConfig == "0001"){
          return "Alarm on closed";
      } elsif($devConfig == "0002"){
          return "Alarm on open";
      } elsif($devConfig == "0100"){
          return "Call on closed";
      } elsif($devConfig == "0200"){
          return "Call on open";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown devConfig value \"$devConfig\"\n";
          return "Unknown value $devConfig";
      }
  }

  if($devType == $TYPE_HOMEKEY){
      my $decode;
      my $top=substr($devConfig, 0, 2);
      my $bot=substr($devConfig, 2, 2);
      #print "top $top / bot $bot\n";
      if($bot == "00"){
          $decode="Silent Alarm";
      } elsif($bot == "01"){
          $decode="Beep Alarm";
      } elsif($bot == "02"){
          $decode="Vibrate Alarm";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown bot value \"$bot\"\n";
          $decode="Unknown value $bot";
      }

      if($top == "00"){
          $decode="$decode / 10 sec battery update";
      } elsif($top == "01"){
          $decode="$decode / 30 sec battery update";
      } elsif($top == "02"){
          $decode="$decode / 1 min battery update";
      } elsif($top == "03"){
          $decode="$decode / 2 min battery update";
      } elsif($top == "04"){
          $decode="$decode / 15 min battery update";
      } elsif($top == "05"){
          $decode="$decode / 1 hour battery update";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown tot value \"$top\"\n";
          $decode="$decode Unknown value $top";
      }
      return $decode;
  }

  if($devType == $TYPE_BASE){
      my $decode;
      my $top=substr($devConfig, 0, 2);
      my $bot=substr($devConfig, 2, 2);
      if($bot == "00"){
          $decode="Global don't call me";
      } elsif($bot == "01"){
          $decode="Global call me always";
      } elsif($bot == "02"){
          $decode="Global automatic call me";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown bot value \"$bot\"\n";
          $decode="Unknown value $bot";
      }

      if($top == "00"){
          $decode="$decode / Do not send system info";
      } elsif($top == "01"){
          $decode="$decode / Send system info once a day";
      } elsif($top == "02"){
          $decode="$decode / Send system info twice a day";
      } elsif($top == "03"){
          $decode="$decode / Send system info four times a day";
      } else {
          print ERRORLOG "$date - devConfigLookup() with devType \"$devType\" had unknown top value \"$top\"\n";
          $decode="$decode / Unknown value $top";
      }
      return $decode;
  }

  if($devType == $TYPE_MODEM){
    return "N/A";
  }

}

sub checkAlert
{
  if($devAlert == "00")
  {
    return "no alert";
  } else {
    if($devAlert == "01"){
        &alertNotify();
        return "alarm triggered"; # controlled by field 9 - $devConfig
    } elsif($devAlert == "04"){
        &alertNotify();
        return "low battery";
    } elsif($devAlert == "08"){
        &alertNotify();
        return "battery charging"; # only valid for home key
    } elsif($devAlert == "20"){
        &alertNotify();
        return "running on backup battery"; # only valid for base station
    } else {
        print ERRORLOG "$date - devAlert() had unknown value \"$devAlert\"\n";
        &alertNotify();
        return "unknown alert";
    }
  }
}

sub alertNotify
{
  if($alert != "00000")
  {
    my $AllAlerts = "";
    my @alertC = split("", $alert);
    if(@alertC[0] == "1"){
      $AllAlerts = "Unknown";
      #my $alertUnknown = "1";
    }
    if(@alertC[1] == "1"){
      $AllAlerts = "$AllAlerts, Base Station running on backup battery";
      #my $alertBackupBatt = "1";
    }
    if(@alertC[2] == "1"){
      $AllAlerts = "$AllAlerts, Home Key battery charging";
      #my $alertBattCharge = "1";
    }
    if(@alertC[3] == "1"){
      $AllAlerts = "$AllAlerts, Low sensor battery";
      #my $alertLowBatt = "1";
    }
    if(@alertC[4] == "1"){
      $AllAlerts = "$AllAlerts, Alarm set";
      #my $alertAlarm = "1";
    }
    $AllAlerts =~ s/^ ,//g;

    use Email::MIME;
    my $message = Email::MIME->create(
      header_str => [
      From    => "$emailFrom",
      To      => "$emailTo",
      Subject => "Home Heartbeat Alert from $devName",
    ],
    attributes => {
      encoding => 'quoted-printable',
      charset  => 'ISO-8859-1',
    },
    body_str => "An alert was detected: $devName signaled $AllAlerts.\n",
    );
    use Email::Sender::Simple qw(sendmail);
    sendmail($message);
    print ALERTLOG "$date - Alert email sent for $devName for alert $AllAlerts";
  }
}
